#!-PERL-
# File:		mif2mif.pl
# Version:	1.3
# Copyright:	MacDonald Dettwiler and Associates, 1994
# Author:	Duncan Fraser (djf@mda.ca)
# This file may be modified provided this header is kept intact
# and all modifications are documented here.
# This code may not be included in a commercial product without
# the permission of the author.

# This script is part of the fm2html package.

# CHANGES
#
# Version 1.3
#   - "-e" flag eliminated since mifren can be used to rename a
#     book or document
#   - "-rm" changed to "-r"
#
# Version 1.2
#   - tags in the book file are now converted, because generated files
#     rely on them
#   - FrameMaker 4.0 book files are now supported
#   - if "-rm" is specified then input and auxiliary files are removed
#     after successful creation of output file
#
# Version 1.1
#   - assumes initially that MIF file in a book has same name as document 
#     but with ".mif" appended; however, if file with that name is not
#     found, looks for name with ".doc" stripped, e.g. given "foo.doc" 
#     looks first for "foo.doc.mif", then for "foo.mif"
#   - by default, files are now modified in-place, e.g. "foo.mif" is moved
#     to "foo.mif.orig" (or "foo.mif.bak" if "foo.mif.orig" already exists)
#     and output is written to "foo.mif"; however, if "-e" option is used
#     to specify a filename extension then output is written to a new
#     filename, e.g. "foo.mif" is left untouched and output is written to
#     "foo.extension.mif"
#   - if "-e" is specified, cross-references are updated to point to new
#     filenames, e.g. to point to "foo.extension.doc" instead of "foo.doc"

$\ = "\n";		# set output record separator
$, = " ";		# set output field separator
$* = 0;			# set single line pattern matching

select( STDERR );	# all error messages go to stderr
$| = 1;			# make stderr unbuffered

$FALSE = 0;
$TRUE  = 1;

$MIF_MAGIC		= "<MIFFile";
$MIF_BOOK_MAGIC		= "<Book ";
$FM_MAGIC		= "<MakerFile";
$FM_BOOK_MAGIC		= "<Bookfile";

$PTAGS_EXTENSION	= "ptags";
$CTAGS_EXTENSION	= "ctags";

$GLOBAL_PTAGS_FILE	= "-LIBDIR-/ptags";
$GLOBAL_CTAGS_FILE	= "-LIBDIR-/ctags";

$TMPNAME		= "mif2mif.tmp";


sub escape	# take string, insert backslash before special characters
{
  local( $_ ) = @_;
  s|([\.\+\?\*\'\"\(\)])|\\$1|g;
  $_;
}

sub dirname	# take pathname, return leading directories
{
  local( $_ ) = @_;
  s|^(.+)/.*$|$1| || s|^/[^/]*$|/| || s|^[^/]*$|.|;
  $_;
}

sub basename	# take pathname, strip leading directories and optional suffix
{
  local( $_, $suffix ) = @_;
  m|/|o && s|^.*/([^/]*)$|$1|;
  $suffix = &escape( $suffix );
  s|(.*)$suffix|$1|;
  $_;
}


local( $prog ) = &basename( $0 );
local( $usage ) = "Usage: $prog [-p ptags_file] [-c ctags_file] [-q] [-r] filename";

local( $tmpdir, $tmpfile, $infile, $indir, $cwd, $outfile, $file,
       $ptags_file, $ctags_file, @mifs, %names, @extras,
       %paths, $base, $g_ptags, $g_ctags );

local( $quiet ) = $FALSE;
local( $remove ) = $FALSE;
local( $exitval ) = 0;

while ( $ARGV[0] =~ m|^-| )
{
  $_ = shift;
  last if m|^--|;
  
  if ( m|^-c$| )
  {
    $ctags_file = "$ARGV[0]"; shift; next;
  }
  elsif ( m|^-p$| )
  {
    $ptags_file = "$ARGV[0]"; shift; next;
  }
  elsif ( m|^-q$| )
  {
    $quiet = $TRUE; next;
  }
  elsif ( m|^-r$| )
  {
    $remove = $TRUE; next;
  }

  print "$prog: ERROR: Invalid flag: \"$_\"";
  print "$usage";
  exit( 1 );
}

if ( $#ARGV != 0 )
{
  print "$prog: ERROR: You must specify exactly one MIF book or document as input";
  print "$usage";
  exit( 1 );
}
  
$infile = "$ARGV[0]";

$quiet == $FALSE && print "";

if ( ! -e "$infile" )
  { die "$prog: ERROR: Input file \"$infile\" does not exist... exiting"; }
if ( ! open( INFILE, "<$infile" ) )
  { die "$prog: ERROR: Can't open input file \"$infile\"... exiting"; }

$base = "$infile";
$base =~ s|\.mif$||;

"$ptags_file" || ($ptags_file = "$base.ptags");
if ( ! -e $ptags_file )
{
  $quiet == $FALSE &&
    print "$prog: $ptags_file not found; copying global ptags file...";
  if ( -e $GLOBAL_PTAGS_FILE )
    { eval system( "cp $GLOBAL_PTAGS_FILE $ptags_file" ); }
  else
    { die "$prog: ERROR: Global ptags file \"$GLOBAL_PTAGS_FILE\" not found... exiting"; }
}
open( PTAGSF, "<$ptags_file" )
  || die "$prog: ERROR: Can't open tags file \"$ptags_file\"... exiting";

$ctags_file || ($ctags_file = "$base.ctags");
if ( ! -e $ctags_file )
{
  $quiet == $FALSE &&
    print "$prog: $ctags_file not found; copying global ctags file...";
  if ( -e "$GLOBAL_CTAGS_FILE" )
    { eval system( "cp $GLOBAL_CTAGS_FILE $ctags_file" ); }
  else
    { die "$prog: ERROR: Global ctags file \"$GLOBAL_CTAGS_FILE\" not found... exiting"; }
}
open( CTAGSF, "<$ctags_file" ) ||
  die "$prog: ERROR: Can't open tags file \"$ctags_file\"... exiting";

if ( $quiet == $FALSE )
{
  print "$prog: Using paragraph format tags from file \"$ptags_file\"";
  print "$prog: Using character format tags from file \"$ctags_file\"";
}

$magic = <INFILE>;

if ( $magic =~ m|^$FM_MAGIC|i || $magic =~ m|^$FM_BOOK_MAGIC|i )
  { die "$prog: ERROR: FrameMaker file \"$infile\" is not a MIF file... exiting"; }
if ( $magic !~ m|^$MIF_MAGIC|i && $magic !~ m|^$MIF_BOOK_MAGIC|i )
  { die "$prog: ERROR: Input file \"$infile\" is not a FrameMaker file... exiting"; }

close( INFILE );
 
# Docs and xrefs may have relative pathnames so it is necessary to chdir.
# Unfortunately, however, FrameMaker sometimes has problems figuring out
# what the current directory is and defaults to the user's home directory,
# in which case it won't find the files anyway.
$indir = &dirname( "$infile" );
chdir( "$indir" ) ||
  die "$prog: ERROR: Can't change directory to \"$indir\"... exiting";
chop( $cwd = `pwd` );
$infile = &basename( "$infile" );
$infile = "$cwd/$infile";

$tmpdir = $ENV{'TMPDIR'};
if ( ! $tmpdir ) { $tmpdir = "-TMPDIR-"; }
if ( ! $tmpdir ) { $tmpdir = "$cwd"; }
$tmpdir =~ s|/$||;
$tmpfile = "$tmpdir/$TMPNAME.$$";

if ( $magic =~ m|^$MIF_BOOK_MAGIC|i )
{
  open( INFILE, "<$infile" ) ||
    die "$prog: ERROR: Can't open input file \"$infile\"... exiting";

  $mifs[++$#mifs] = "$infile";

  while ( <INFILE> )
  {
    if ( m|<FileName | )
    {
      local( $docname, $docpath, $oldname, $mifpath );

      if ( m|<FileName \s*`<c\\>([^<]*)<U\\>([^']*)'>| )
      {
        # FrameMaker 3.1
        $docname = "$1";
        $docpath = "$2";
      }
      elsif ( m|<FileName \s*`<c\\>([^<]*)'>| )
      {
        # FrameMaker 4.0
        $docname = "$1";
        $docpath = "$1";
      }
      
      $oldname = "$docname";
      $mifpath = "$docpath.mif";

      if ( ! -e "$mifpath" )
      {
	local( $unstripped ) = "$mifpath";

	$mifpath =~ s|\.doc||;

	if ( $mifpath eq $unstripped )
	{
	  print "$prog: ERROR: Input file \"$unstripped\" does not exist";
	  $exitval = 1;
	  next;
	}
	elsif ( ! -e $mifpath )
	{
	  print "$prog: ERROR: Input file \"$unstripped\" does not exist, neither does \"$mifpath\"";
	  $exitval = 1;
	  next;
	}
      }

      $mifs[++$#mifs] = "$mifpath";
      $names{$oldname} = "$docname";
      $paths{$oldname} = "$docpath";
    }
  }

  close( INFILE );

  @ARGV = @mifs;
}

open( TMPF, ">$tmpfile" ) || die "Can't open temporary file \"$tmpfile\"";

print TMPF <<END;
#!-PERL-

select( STDERR );
\$| = 1;

\$FALSE = 0;
\$TRUE  = 1;

sub substitute
{
  local( \$inf, \$outf ) = \@_;
  local( \$magic, \$\\ );

  \$\\ = \"\\n\";

  if ( ! -e \$inf )
  {
    print \"$prog: ERROR: Can't convert \\\"\$inf\\\" (does not exist)\";
    return 1;
  }

  if ( ! open( INF, \"<\$inf\" ) )
  {
    print \"$prog: ERROR: Can't open file \\\"\$inf\\\"\";
    return 1;
  }

  if ( ! open( OUTF, \">\$outf\" ) )
  {
    print \"$prog: ERROR: Can't create file \\\"\$outf\\\"\";
    return 1;
  }

  \$magic = <INF>;

  if ( \$magic !~ m|$MIF_MAGIC|i && \$magic !~ m|$MIF_BOOK_MAGIC|i )
  {
    print \"$prog: ERROR: File \\\"\$inf\\\" is not a MIF file\";
    return 1;
  }

  chop( \$magic ); print OUTF \$magic;
  print OUTF "\n# This file has been modified by version 1.3 of the mif2mif filter\n";

  while ( <INF> )
  {
END
  
while ( <PTAGSF> )
  { m|^([^\t]+)\t+([^\t\n]+)[\n\t]|o && ($1 ne $2) && ($g_ptags{$1} = "$2"); }
close( PTAGSF );

print TMPF <<END;
    if ( m|<PgfTag|o || m|<DeriveTag|o )
    {
END

while ( ($from, $to) = each( %g_ptags ) )
{
  local( $escaped ) = &escape( $from );
  print TMPF "      s|`$escaped'|`$to'|o;";
}

print TMPF <<END;
    }
    elsif ( m|<XRefSrcText|o || m|<MText|o )
    {
END

while ( ($from, $to) = each( %g_ptags ) )
{
  local( $escaped ) = &escape( $from );
  print TMPF "      s|: $escaped:|: $to:|o;";
}

print TMPF <<END;
    }
    elsif ( m|<XRefSrcFile|o )
    {
END

while ( ($from, $to) = each( %names ) )
{
  local( $escaped ) = &escape( $from );
  print TMPF "      s|`<c\\\\>$escaped'|`<c\\\\>$to'|o;";
}

print TMPF <<END;
    }
    elsif ( m|<FileName |o )
    {
END

while ( ($from, $to) = each( %names ) )
{
  local( $escaped ) = &escape( $from );
  local( $to_path ) = "$paths{$from}";
  # FrameMaker 3.1
  print TMPF "      s|`<c\\\\>$escaped<U\\\\>[^']*'|`<c\\\\>$to<U\\\\>$to_path'|o;";
  # FrameMaker 4.0
  print TMPF "      s|`<c\\\\>$escaped'|`<c\\\\>$to'|o;";
}

print TMPF <<END;
    }
    elsif ( m|<VariableDef `<\\\$paranum\\[|o )
    {
END

while ( ($from, $to) = each( %g_ptags ) )
{
  local( $escaped ) = &escape( $from );
  print TMPF "      s|\\[$escaped\\]|[$to]|o;";
}

print TMPF "    }";

while ( <CTAGSF> )
  { m|^([^\t]+)\t+([^\t\n]+)[\n\t]|o && ($1 ne $2) && ($g_ctags{$1} = "$2"); }
close( CTAGSF );

if ( %g_ptags )
  { print TMPF "    elsif ( m|<FTag|o )"; }
else
  { print TMPF "    if ( m|<FTag|o )"; }

print TMPF "    {";

while ( ($from, $to) = each( %g_ctags ) )
{
  local( $escaped ) = &escape( $from );
  print TMPF "      s|`$escaped'|`$to'|o;";
}

print TMPF "    }";

print TMPF <<END;

    chop; print OUTF;
  }

  0;	# successful return
}
END

close( TMPF );
require( "$tmpfile" );


while ( $file = shift )
{
  local( $retval );

  $infile  = ( -e "$file.orig" ) ? "$file.bak" : "$file.orig"; 
  $outfile = "$file";
  
  $quiet == $FALSE && print "$prog: Renaming \"$file\" as \"$infile\"...";
  
  if ( ! rename( "$file", "$infile" ) )
  {
    print
      "$prog: ERROR: Can't convert \"$file\" (can't rename as \"$infile\")";
    $exitval = 1;
    next;
  }

  $extras[++$#extras] = "$infile";

  $quiet == $FALSE &&
    print "$prog: Converting \"$infile\" to \"$outfile\"...";
  $retval = &substitute( $infile, $outfile );
  $retval != 0 && ($exitval = $retval);

  if ( $remove == $TRUE )
  {
    $quiet == $FALSE && print "$prog: Deleting original file \"$infile\"...";
    eval system( "rm -f $infile" );
  }
}

eval system( "rm -f $tmpfile" );

if ( $remove == $TRUE )
{
  $quiet == $FALSE && print
    "$prog: Deleting tags files \"$ptags_file\" and \"$ctags_file\"...";
  eval system( "rm -f $ptags_file $ctags_file" );
}

$quiet == $FALSE && print "$prog: Done\n";

exit( $exitval );
